#!/bin/sh
# \
exec tcl "$0" "$@"


#								 ***********
#								** Summary **
#								 ***********

#					  Copyright 2002, Nagual Enterprises
#				Please see COPYING for the complete GNU license

#	(1)	Title:
#				Chunk Converter

#	(2)	Purpose:
#			To convert the byte stream of a Daggerfall saved game
#			into proper Tcl lists for machination.

#	(3)	Platform dependencies:
#			Left-handed chirality is assumed.

#	(4)	Change history:
#			Please see the RCS logs for change history.

#	(5)	Definitions:

#			Namespaces:
#				Arena2SaveTree		--	Chunk extract/update class
#				Arena2DataHeader	--	Form common to all chunks

#			Exports:
#				chunkName			--	Answer the descriptive name for
#										a chunk
#				chunksFor			--	Answer the collection of chunks
#										for a Daggerfall saved game
#				contents			--	Insert the formated representation
#										of a chunk collection into a listbox
#				helpText			--	Answer the status bar display for a
#										chunk header field
#				hpHardness			--	Answer the point spread for the
#										hp per level setting
#				listString			--	Format the description of a chunk
#										as a listbox entry
#				name2Key			--	Answer the chunk type from the
#										text description of a chunk
#				names				--	The descriptive names of all the
#										chunks
#				player				--	The instance of the player object
#				printList			--	Format the chunk header in the
#										style `keyName: keyValue'
#				printValues			--	Format just the values of the chunk
#										header fields
#				raw2Tcl				--	Convert a chunk header byte stream to
#										a proper Tcl list
#				size				--	The (fixed) chunk header size
#				tcl2Raw				--	Convert a Tcl list to a saved game
#										byte stream
#				updatePlayer		--	Update the saved game image of the
#										player chunk
#				widgetList			--	Answer the widget requirements for
#										displaying the fields of a chunk
#										header

#			Assertions about definitions:
#				***  None  ***

#	(6)	References:

#			Library files:
#				Tclx				--	Extended Tcl support

#			Imports:
#				None, but see Arena2DataHeader imports for convenience.

#			Assertions about the references:
#				***  None  ***

#								 ********
#								** Text **
#								 ********

#							 *****************
#							** Configuration **
#							 *****************

namespace eval Arena2SaveTree {
	variable exeDir				"/dos/g/game/dag"
	variable artDir				"/dos/g/game/dag/arena2"
	variable tmpDir				"/tmp"
	variable saveTcl			"stree.tcl"
	variable saveTree			"savetree.dat"
	variable saveName			"savename.txt"
	variable saveImage			"image.raw"
	variable palette			"art_pal.col"
	variable playerLine			3
	variable playerOffset		7
	variable playerToken		20
	variable playerSize			705
	variable reversePatches		0
	variable skillOffset		228
	variable computedDifficulty	65536
	variable save0
	variable save1
	variable save2
	variable save3
	variable save4
	variable save5

	namespace export chunksFor
	namespace export contents
	namespace export player
	namespace export hpHardness
	namespace export updatePlayer
}

namespace eval Arena2DataHeader {
	variable hasExpand			1
	variable keys	{ offset key filler0a x y z filler0b filler0c picture1 picture2
					  objectId  filler1a questNo parentId filler1 itemObject
					  questObject nextObject childObject sublistHead parentKey }
	variable keyHelp {	{Offset in savetree.dat}
						{Type of chunk}
						{Unknown}
						{Cartesian x}
						{Cartesian y}
						{Cartesian z}
						{Unknown}
						{Unknown}
						{3D view picture}
						{Inventory picture}
						{Chunk identifier}
						{Unknown}
						{Quest number}
						{Parent chunk identifier}
						{Unknown}
						{Active spell/spell book/permanent treasure container}
						{Quest object identifier}
						{Link to next object in series}
						{Link to first child object in series}
						{Link to parent object of series}
						{Chunk type of parent}}
	variable keyLinks	{ sort sort text text text text text text image image sort
						  text sort link text sort link sort
						  sort sort sort }
	variable print	{ %08x %02x %02hx%02hx%02hx   %08x %08x %08x   %08x %08x %04hx %04hx %08x
					  %02hx%02hx%02hx %02x %08x %08x %08x %08x %08x
					  %08x %08x %08x }
	variable scan	"c1 c6 iii ii ss i c3 c1 i ii ii iii"
	variable size	71
	variable links	{ {parentId objectId} {questObject objectId} }

	variable names	{
		"Item Auxiliary" "World" "Item" "Player"
		"Move" "Person" "Unknown_06" "Underground?"
		"Unknown_08" "Spell" "Guild" "Original Stats"
		"" "" "Qbn image" ""
		"Quest Holder" "" "Monster" ""
		"" "" "Spellcaster" "Control panel"
		"Logbook" "Bank accounts" "" ""
		"Original class" "Old Guild" "" "Potion"
		"Door" "Player discards" "Creature" ""
		"Store item" "" "" "NonWorld"
		"Quest site" "Quest NPC" "Child guard" "Unknown_2b"
		"Corpse" "NPC" "Generic NPC" ""
		"" "" "" "Current dungeon"
		"Container" "Townsperson" "" ""
		"" "" "Treasure pile" ""
		"" "" "" ""
		"Visited?" "Qbn questor" "" ""
	}
	namespace export names
	namespace export size
	namespace export chunkNane
	namespace export helpText
	namespace export listString
	namespace export name2Key
	namespace export raw2Tcl
	namespace export tcl2Raw
	namespace export printList
	namespace export printValues
	namespace export widgetList
}
namespace import -force Arena2DataHeader::label
namespace import -force Arena2DataHeader::name2Key
namespace import -force Arena2DataHeader::printList
namespace import -force Arena2DataHeader::raw2Tcl
namespace import -force Arena2DataHeader::tcl2Raw


#							 ********************
#							** Instance methods **
#							 ********************

proc Arena2DataHeader::chunkName {self} {
#		Answer the chunk name for self.

	variable names
	return [lindex $names [keylget self "key"]]
}

proc Arena2DataHeader::helpText {key} {
	variable keyHelp
	variable keys
	set index [lsearch $keys $key]
	if { $index >= 0 } then {
		return [lindex $keyHelp $index]
	}
	return ""
}

proc Arena2DataHeader::listString {self} {
#		Answer the descriptive label for self.

	variable hasExpand
	set name [keylget self "name"]
	if { "$name" == "" } then {
		set name [format 0x%02x [keylget self "key"]]
	}
	set oid [keylget self "objectId"]
	set aux ""
	switch -- [format %#x $oid] {
		0x12715 {set aux " (Weapons)"}
		0x12716 {set aux " (Magic items)"}
		0x12717 {set aux " (Clothes, etc)"}
		0x12718 {set aux " (Ingredients)"}
		0x12719 {set aux " (Small boat)"}
		0x1271a {set aux " (Large boat)"}
	}
	if {$hasExpand} then {
		set reply [format "%08x %s$aux" $oid $name]
#		set reply [exec echo "$reply" | expand]
	} else {
		set reply [format "%08x %s$aux" $oid $name]
	}
	return $reply
}

proc Arena2DataHeader::name2Key {text} {
#		Translate a chunk text description to a type number.

	variable names
	set lower [string tolower "$text"]
	set index 0
	foreach name $names {
		if {[string match "$lower*" [string tolower "$name"]]} {
			return $index
		}
		incr index
	}
	return -1
}

proc Arena2DataHeader::printList {self} {
#		Print a formated copy of self.

	variable keys
	variable print
	set reply {}
	foreach key $keys format $print {
		set data [keylget self $key]
		set cmd [concat format \"%16.16s: $format\\n\" $key $data]
		set ans [eval $cmd]
		lappend reply $ans
	}
	return $reply
}

proc Arena2DataHeader::printValues {self} {
#		Print a formated copy of self.

	variable keys
	variable print
	set reply {}
	foreach key $keys format $print {
		set data [keylget self $key]
		set cmd [concat format \"$format\\n\" $data]
		set ans [eval $cmd]
		lappend reply $ans
	}
	return $reply
}

proc Arena2DataHeader::raw2Tcl {fid} {
#		Convert the raw byte stream on fid to a Tcl list.

	variable keys
	variable scan
	variable size

	set offset [tell $fid]
	set data [read $fid $size]
	eval binary scan "\$data" "\$scan" [join [lrange $keys 1 end]]
	set reply {}
	keylset reply offset $offset
	foreach one $keys {
		set field {}
		eval set field [cconcat \$ $one]
		keylset reply $one $field
	}
	set keyName [chunkName $reply]
	keylset reply "name" $keyName
	return $reply
}

proc Arena2DataHeader::tcl2Raw {self fid} {
#		Convert the Tcl list in self to a raw byte
#		stream on fid.

	variable keys
	variable scan
	variable size

	set data {}
	foreach key $keys {
		lappend data [keylget self $key]
	}
	set data [concat $data]
	set length [clength $data]
	if {$length != $size} then {
		puts stderr "Error: Arena2DataHeader size ($length) is not $size"
		return 0
	}
	keylget self "offset" offset
	seek $fid $offset start
	puts -nonewline $fid [binary format $scan [concat $data]]
	return 1
}

proc Arena2DataHeader::widgetList {} {
#		Print a formated copy of self using the widgets for
#		its varient and invarient display parts.

	variable keys
	variable keyLinks
	set lines {}
	foreach key $keys link $keyLinks {
		lappend lines [list $link "$key"]
	}
	return $lines
}

proc Arena2SaveTree::allChunks { path } {
#		Open the specified save game and collect the chunk record.

	variable chunkSize
	variable playerOffset
	variable savePosition
	variable ::Arena2DataHeader::size
	set fid [openChunk $path]
	set fields {}
	set data {}
	while {[nextChunk $fid]} {
		set chunk [Arena2DataHeader::raw2Tcl $fid]
		keylset data [keylget chunk "offset"]  [read $fid [expr $chunkSize - $::Arena2DataHeader::size]]
		lappend fields $chunk
	}
	::close $fid
	return [list $fields $data]
}

proc Arena2SaveTree::atPut { field data } {
#		Update the binary data field with fresh information converted
#		converted from ascii data.

	variable reversePatches

	keylget field "size" length
	set firstChar [ctype digit [cindex $data 0]]
	set relative ! [ctype digit $firstChar]
	switch -- $length {
		1	-
		2	-
		4	{
#			Update a numeric field (byte, short, or long)

			keylget field "ascii" oldValue
			if {$reversePatches} {
				set newValue [expr - $oldValue]
			} else {
				set newValue $data
			}
			if {$relative} {
				set newValue [expr $newValue + $oldValue]
			}
			keylset field "ascii" $newValue
			keylset field "data" [binary format [cindex " cs i" $length] $newValue]
		}
		-1025	-
		-2049	-
		-3073	-
		-4097	-
		-5121	-
		-513	{
#			Update mana x Int.
 
			keylget field "ascii" oldValue
			if {$firstChar == y || $firstChar == Y || $firstChar == 1} {
				set newValue 1
			} else {
				set newValue 0
			}
			if {$reversePatches} {
				set newValue !$newValue
			}
			set length [expr [expr - $length] >> 8]
			set oldValue [expr $oldValue & ~0x1c]
			if {$length == 2} {
				set length 0
			}
			if {$newValue} {
				set oldValue [expr $oldValue | $length]
			} else {
				set oldValue [expr $oldValue | 0x14]
			}
			keylset field "ascii" $oldValue
		}
		-1		-
		-2		-
		-4		-
		-8		-
		-16		-
		-32		-
		-64		-
		-128	-
		-256	-
		-512	-
		-1024	-
		-2048	-
		-3072	-
		-4096	-
		-8192	{
#			Update a shared bit field.

			keylget field "ascii" oldValue
 			if {$firstChar == y || $firstChar == Y || $firstChar == 1} {
 				set newValue 1
 			} else {
 				set newValue 0
 			}
 			if {$reversePatches} {
 				set newValue !$newValue
 			}
 			if {$length < -128} {
 				set length [expr [expr - $length] >> 8]
 			} else {
 				set length [expr - $length]
 			}
 			if {$newValue} {
 				set oldValue [expr $oldValue | $length]
 			} else {
 				set oldValue [expr $oldValue & ~$length]
 			}
 			keylset field "ascii" $oldValue
		}
		default {
#			Update a string value.

			keylset field "ascii" $data
			keylset field "data" [binary format "a*x" $data]
		}
	}
}

proc Arena2SaveTree::chunksFor {saveId} {
#		Collect the chunks in the specified saved game.  saveId can
#		be either the saved game slot number in [0,5] or the saved
#		game subdirectory name: save0, save1, ... save5.

	variable exeDir
	variable saveTree

	if {[ctype digit $saveId]} then {
		set savetree [file join $exeDir "save$saveId" "$saveTree"]
	} else {
		set savetree [file join $exeDir "$saveId" "$saveTree"]
	}
	return [allChunks $savetree]
}

proc Arena2SaveTree::contents {chunks pane} {
#		Fill the contents of the list pane with the chunk names
#		in the receiver.

	$pane delete 0 end
	foreach chunk $chunks {
		$pane insert end [Arena2DataHeader::listString $chunk]
	}
}

proc Arena2SaveTree::developedSkills {aList} {
#		Assign the developed skill sets: primary, major, and minor
#		as indices to aList.

	upvar $aList record
	foreach key {primary1 primary2 primary3 major1 major2 major3
				 minor1 minor2 minor3 minor4 minor5 minor6} {
		keylget record $key values
		keylget values "ascii" skill
		set skillField [lindex $record $skill]
		set values [lindex $skillField end]
		keylset values "skill" $key
		set skillField [lreplace $skillField end end $values]
		set record [lreplace $record $skill $skill $skillField]
	}
}

proc Arena2SaveTree::hpHardness aList {
#		Answer the difficulty factor for the current hit points per level."

	upvar $aList player
	keylget player "hpPerLevel" field
	keylget field "ascii" hpLevel
	if {$hpLevel >= 8} {
		return [expr 3277 * ($hpLevel - 8)]
	} else {
		return [expr -6554 * (8 - $hpLevel)]
	}
}

proc Arena2SaveTree::init { aList recordData field } {
#		Answer the record field list at a field name in aList by
#		populating the "ascii" field from its present binary value
#		upon the first access to the field.

	variable computedDifficulty
	keylget field "hard" hard
	keylget field "size" length
	switch -- $length {
		0	{
			keylset field "ascii" -1
		}
		1	-
		2	-
		4	{
#			Retrieve a numeric field (byte, short, or long)

			keylget field "offset" offset
			set value [csubstr $recordData $offset $length]
			binary scan $value [cindex " cs i" $length] ascii
			keylset field "ascii" $ascii
			set computedDifficulty [expr $computedDifficulty + $hard]
		}
 		-1025	-
 		-2049	-
 		-3073	-
 		-4097	-
 		-5121	-
 		-513	{
 #			Retrieve mana x Int.
 
			set value [csubstr $recordData [keylget field "offset"] 2]
			binary scan $value [cindex " cs i" 2] ascii
			set manaPerInt [expr $ascii >> 10]
			set length [expr [expr - $length] - 1]
			if {$length == 512} {
				set length 0
			}
			set mask [expr $ascii & 0x1c00]
			if {$mask == $length} {
				keylset field "ascii" 1
				set computedDifficulty [expr $computedDifficulty + $hard]
			} else {
				keylset field "ascii" 0
			}
		}
		-1		-
		-2		-
		-4		-
		-8		-
		-16		-
		-32		-
		-64		-
		-128	-
		-256	-
		-512	-
		-1024	-
		-2048	-
		-3072	-
		-4096	-
		-8192	{
#			Retrieve a shared bit field.

			set value [csubstr $recordData [keylget field "offset"] 2]
			binary scan $value [cindex " cs i" 2] ascii
			keylget field "mask" mask
			set length [expr - $length]
			set mask [expr $ascii & $length]
#			set mask [expr $ascii & $mask]
			if {$mask} {
				keylset field "ascii" 1
				set computedDifficulty [expr $computedDifficulty + $hard]
			} else {
				keylset field "ascii" 0
			}
		}
		default {
#			Retrieve a string value.

			set value [csubstr $recordData [keylget field "offset"] $length]
			binary scan $value A* ascii
			keylset field "ascii" $ascii
		}
	}
	return $field
}

proc Arena2SaveTree::nextChunk { fid } {
#		Process the next savetree.dat chunk.  Answer false if there is
#		no next chunk, and true otherwise.

	variable chunkSize
	variable savePosition
	variable empty
	if { [eof $fid] } then {
		return 0
	}
	set savePosition [expr $savePosition + $chunkSize + 4]
	seek $fid $savePosition start
	if { [eof $fid] } then {
		return 0
	}
	while {1} {

#		This is arguably wrong, but hopefully we will never see a saved
#		game approaching 2 gigabytes in size.

		binary scan [read $fid 4] i chunkSize
		if {$chunkSize != 0 && $empty > 0} then {
			return 1
		}
		set savePosition [expr $savePosition + $chunkSize + 4]
		seek $fid $savePosition start
		incr empty
		if {$empty >= 3} then {
			break
		}
	}
	return 0
}

proc Arena2SaveTree::openChunk { path } {
#		Open the savetree file specified by a path.  Answer the
#		opened file handle.

	variable chunkSize
	variable playerOffset
	variable savePosition
	variable empty
	set chunkSize 0
	set playerOffset 0
	set savePosition 0x0f
	set fid [::open $path ]
	fconfigure $fid -translation binary
	seek $fid 0x13 start
	set empty 0
	return $fid
}

#eval $argv
