#!/bin/sh
# \
exec wishx "$0" "$@"


#								 ***********
#								** Summary **
#								 ***********

#					  Copyright 2002, Nagual Enterprises
#				Please see COPYING for the complete GNU license

#	(1)	Title:
#				Hex Dump Pane

#	(2)	Purpose:
#			To manage the display and update of a hexadecimal
#			dump-style text widget.

#	(3)	Platform dependencies:
#			Note that `set' is presently broken within a namespace eval {}
#			construct so we gratuitously pollute the namespace with
#			temporary variables.  This is probably induced by the
#			(hokey) alias mechanism, as direct invocations seemed not
#			to complain at all.

#	(4)	Change history:
#			Please see the RCS logs for change history.

#	(5)	Definitions:

#			Namespaces:
#				DumpView			--	Class controller to manage the
#										display of a hexadecimal stream,
#										rather than a character stream,
#										inside a text widget

#			Exports:
#				columns				--	Answer the number of bytes per row
#										in the display
#				columns:			--	Assign the number of bytes per row
#										in the display
#				contents			--	Answer the formatted collection of
#										lines in the receiver
#				contents:			--	Create the formatted collection of
#										lines in the receiver from a byte
#										stream
#				discard				--	Permanently retire an instance
#				font				--	Answer the current display font
#				font:				--	Assign the display font to use
#				new					--	Creates an independent clone of
#										the DumpView widget
#				openInside			--	Open the receiver's text widget
#										inside a top pane
#				rawContents			--	Answer the raw byte data as a
#										Tcl list using hex encoding

#			Assertions about definitions:
#				The display font must be fixed or monospaced for the
#				hex bytes to align visually.

#				Packing the widget into its top pane expands and
#				fills both Cartesian directions.

#				`openInside' will spawn independent clones of
#				the widget and answer the name of the independent
#				widget, which can be used for subsequent invocations.
#				E.g.

#					[DumpView::openIn aTopPane] contents: aByteStream

#				The encapsulation isn't perfect: the initial color
#				scheme is read directly from the ChunkView class.


#	(6)	References:

#			Library files:
#				Tclx				--	Extended Tcl support
#				Tk					--	Base widget classes

#			Imports:
#				***  None  ***

#			Assertions about the references:
#				***  None  ***

#								 ********
#								** Text **
#								 ********

namespace eval DumpView {
	variable _counter		0
	variable cellFormat		"%02x "
	variable _columns		16
	variable _dialog
	variable _font			"-Misc-Fixed-Bold-R-Normal--14-100-110-*-C-*-ISO8859-1"
	variable _bytes
	variable _text
	variable _filler
	variable _cellWidth

	namespace export asLines
	namespace export columns
	namespace export columns:
	namespace export contents
	namespace export contents:
	namespace export discard
	namespace export estimateWidth
	namespace export font
	namespace export font:
	namespace export new
	namespace export openInside
	namespace export rawContents
	namespace export widget
}

#	Tune the fonts for the (broken) Windows port

if {"$tcl_platform(platform)" == "windows"} then {
	variable ::DumpView::_font fixedsys
}

proc DumpView::asLines {self columns} {
#		Format the receiver's display in lines of
#		columns-many character dumps, typically 16.
#		columns can be odd, but even looks much better.

	namespace eval $self {
		set index 0
		set start 0
		set length [llength $_bytes]
		if {$length < 1} {
			return
		}
		set half [uplevel expr (\$columns + 1) / 2]
		set pad 0
		$_dialog.pane delete 1.0 end
		while {$index < $length} {
			set fin [expr $index + $half - 1]
			$_dialog.pane insert end [join [lrange $_bytes $index $fin] {}]
			set next $fin
			incr next
			set fin [expr $fin + $half]
			if {$next > $length} {
				set pad [expr $next - $length]
				$_dialog.pane insert end [replicate "$_filler" $pad]
			}
			$_dialog.pane insert end "- "
			$_dialog.pane insert end [join [lrange $_bytes $next $fin] {}]
			variable next $fin
			incr next
			if {$next > $length} {
				set pad [expr $next - ($length + $pad)]
				$_dialog.pane insert end [replicate "$_filler" $pad]
			}
			$_dialog.pane insert end "   "
			$_dialog.pane insert end [join [lrange $_text $index $fin] {}]
			$_dialog.pane insert end "\n"
			set index $next
		}
	}
}

proc DumpView::columns {self} {
#		Answer the number of columns in the hex dump.

	namespace eval $self {
		return $_columns
	}
}

proc DumpView::columns: {self columns} {
#		Assign the number of columns to display in the hexpane.
#		If the new number of columns differs from the original,
#		layout the pane contents again.

	namespace eval $self {
		if {$_columns != $columns} {
			variable _columns $columns
			asLines $self $columns
		}
	}
}

proc DumpView::contents {self} {
#		Answer the current contents of the dump pane.

	namespace eval $self {
		return $_dialog.pane get 1.0 end
	}
}

proc DumpView::contents: {self bytes} {
#		Format the contents of bytes as a hexadecimal dump, with
#		both hex and ascii display fields.  Display the results on
#		the dump pane.

	namespace eval $self {
		variable _bytes {}
		variable _text {}
		set size [uplevel clength "\$bytes"]
		set index 0
		while {$index < $size} {
			set byte [uplevel crange "\$bytes" $index $index]
			binary scan $byte c code
			if {$code < 0} then {
				set code [expr 256 + $code]}
			lappend _bytes [format "$cellFormat" $code]
			if {[ctype print "$byte"]} then {
				lappend _text "$byte"
			} else {
				lappend _text "."
			}
			incr index
		}
		asLines $self $_columns
	}
}

proc DumpView::discard {self} {
#		Permanently remove an instance and its references.

	namespace delete $self
}

proc DumpView::estimateWidth {self} {
#		Estimate the width of the receiver's text pane
#		needed for the requested number of columns.

	namespace eval $self {
		variable _cellWidth [clength [format "$cellFormat" 17]]
		variable _filler [replicate " " $_cellWidth]
		return [expr ($_cellWidth * $_columns) + 7 + $_columns]
	}
}

proc DumpView::font: {self fontName} {
#		Assign fontName as the (fixed) font to render the
#		contents of the dump pane.  Ignored once the pane
#		has been opened.

	namespace eval $self {
		variable _font "$fontName"
	}
}

proc DumpView::font {self} {
#		Answer the current font specification.

	namespace eval $self {
		return $_font
	}
}

proc DumpView::new {topPane} {
#		Create a new data space instance of the receiver's
#		class.  Answer the new data space.

	variable _counter
	variable cellFormat
	variable _columns
	variable _font

	set name DumpView$_counter
	set _dialog $topPane.hexPane$_counter
	incr _counter
	namespace eval :: namespace eval $name \
		[list variable \
			_columns "$_columns" \
			_dialog "$_dialog" \
			_font "$_font" \
			cellFormat \"$cellFormat\" \
			self ::$name \
		]
	set name ::$name
	namespace eval $name {
		namespace import ::DumpView::*
	}
	return $name
}

proc DumpView::openInside {topPane} {
#		Open an instance of the receiver inside a topPane.

	namespace eval [new $topPane] {
		frame $_dialog
		text	$_dialog.pane \
				-width [estimateWidth $self] \
				-height 11 \
				-yscrollcommand "$_dialog.scroll set" \
				-setgrid true \
				-font "$_font"
		scrollbar $_dialog.scroll \
				-troughcolor $::ChunkView::scrollBackground \
				-background $::ChunkView::buttonActiveBackground \
				-activebackground $::ChunkView::buttonActiveForeground \
				-command "$_dialog.pane yview"
		pack	$_dialog.scroll \
				-side right \
				-fill y
		pack	$_dialog.pane \
				-side left \
				-expand yes \
				-fill both
		pack	$_dialog \
				-expand yes \
				-fill both
		return $self
	}
}

proc DumpView::rawContents {self} {
#		Answer the unformated contents of the dump pane as a
#		regular Tcl list.

	namespace eval $self {
		return $_bytes
	}
}

