#!/bin/sh
# \
exec wishx "$0" "$@"

#								 ***********
#								** Summary **
#								 ***********

#					  Copyright 2002, Nagual Enterprises
#				Please see COPYING for the complete GNU license

#	(1)	Title:
#				Daggerfall Saved Game Chunk Viewer

#	(2)	Purpose:
#			To provide a GUI wrapper around the saved game file
#			format for Daggerfall.

#	(3)	Platform dependencies:
#			***  None  ***

#	(4)	Change history:
#			Please see the RCS logs for change history.

#	(5)	Definitions:

#			Namespaces:
#				Arena2Config
#				ChunkView

#			Exports:
#				canInvoke			--	Answers true if creating a new
#										widget is possible at this time
#				dialog				--	Build and display the widget
#										tree of the receiver
#				discard				--	Permanently retire an instance
#				findKey				--	Method to prompt for and locate a
#										chunk by chunk key value
#				findOid				--	Methodto prompt for and locate a
#										chunk by object ID
#				invoke				--	Method to create a fresh instance
#										of the receiver.
#				jumpTo				--	Method to scroll the chunk list box
#										to a hot-linked target
#				nextKey				--	Method to scroll the chunk list box
#										to the next searched-for entry
#				priorKey			--	Method to scroll the chunk list box
#										to the previous searched-for entry
#				showStatus			--	Method to update the status bar
#				sortBy				--	Method to (re)sort the list of chunks
#										from the current hot-link

#			Assertions about definitions:
#				The complete command line invocation syntax is shown
#				below:

#					listview.tcl  [Daggerfall installation]

#				`Daggerfall installation' is the path to a Daggerfall
#				installation directory (e.g. c:\dagger).

#				If no arguments are provided, the save games are sought
#				in the current working directory.

#				If the BLT package is installed, its busy-wait cursor
#				support will be used.  Otherwise, no busy-wait cursor
#				management will occur.

#				The ChunkView class is fully reentrant and reusable.
#				Calls to `invoke' will create separate toplevel
#				windows of the widget hierarchy within the application.

#				There is no separate locking for the animation thread,
#				which relies on the underlying Tcl implementation for
#				locking via the `after' command.  There is a window of
#				opportunity for `clearAnimation' to miss out on the
#				current animation step if Tcl isn't thread-safe.

#	(6)	References:

#			Library files:
#				BLT					--	Optional busy-wait cursor support
#				Tclx				--	Extended Tcl support
#				Tk					--	Base display widget classes

#			Imports:
#				chunk.tcl			--	Transform saved game to Tcl list
#				dumpview.tcl		--	Widget to display bytes-as-hex
#				prompter.tcl		--	Widget to prompt for input
#				texture.tcl			--	Texture file accessor and converter

#			Assertions about the references:
#				prompter.tcl is included only if Tk doesn't already have
#				support for it.

#								 ********
#								** Text **
#								 ********

#							 *****************
#							** Configuration **
#							 *****************

#	Define the configuration space inherited by all the Tcl classes
#	in the application.  Note that the command line will override
#	the settings for artDir and exeDir upon requested.

namespace eval Arena2Config {
#	Directory containing the game textures:
	variable artDir					/dos/g/game/dag/arena2
#	Directory containing the game executables and saved games:
	variable exeDir					/dos/g/game/dag
#	Currently unused
	variable frameRate				300
#	Display monitor gamma correction:
	variable gamma					1.75
#	Directory used for temporary art files:
	variable tmpDir					./
}

#	Import packages required by the application.

#package require Tix
package require Tclx
#package require Tkx
#load libtix.so

#	Import the `busy' cursor command from blt, if available.  Otherwise
#	supply a no-op method.

if {[catch {package require BLT}]} then {
	proc busy {opt pane} {
		return
	}
} else {
	namespace import blt::busy
}

set clicount [string length $argv]
if {$clicount > 1} then {
#	Override the configuration from the cli.

	variable Arena2Config::artDir [lindex $argv 0]/arena2
	variable Arena2Config::exeDir [lindex $argv 0]
}

#	Include the Daggerfall-specific Tcl appendages.

source chunk.tcl
	namespace import -force Arena2SaveTree::chunksFor
	namespace import -force Arena2DataHeader::helpText
	namespace import -force Arena2DataHeader::listString
	namespace import -force Arena2DataHeader::name2Key
	namespace import -force Arena2DataHeader::printList
	namespace import -force Arena2DataHeader::printValues
	namespace import -force Arena2DataHeader::widgetList

source dumpview.tcl
	namespace import -force DumpView::openInside
	namespace import -force DumpView::contents:
	namespace import -force DumpView::rawContents

source texture.tcl
	namespace import -force Arena2Texture::animate
	namespace import -force Arena2Texture::clearAnimation
	namespace import -force Arena2Texture::currentRate
	namespace import -force Arena2Texture::currentRate:
	namespace import -force Arena2Texture::imagesAt
	namespace import -force Arena2Texture::shortAt

source savelist.tcl
	namespace import -force Arena2SaveGame::saveGameDialog

#	Decide whether we need prompt.tcl or not.

if {[catch {set havePrompter "$auto_index(tk_prompter)"}]} then {
	source prompter.tcl
}

#	Create the application namespace.

namespace eval ChunkView {
	variable applicationBackground	#8989a1
	variable background				ShellBackground
	variable buttonActiveBackground	MenuTopShadow
	variable buttonActiveForeground	PaleTurquoise1
	variable buttonBackground		IconForeground
	variable buttonForeground		White
	variable checkColor				gold
	variable canAnimate				[file executable txttoppm]
	variable columnBackground		ShellBackground
	variable columnForeground		gold
	variable computeTrough			gold2
	variable _counter				0
	variable daggerTrough			SteelBlue
	variable editBackground			MidnightBlue
	variable editDisabledBackground	SteelBlue
	variable editDisabledForeground	White
	variable editForeground			White
	variable extraSpace				0
	variable field					{}
	variable foreground				White
	variable gamma					2.4
	variable legendBackground		SteelBlue
	variable listBackground			EditBackground
	variable listFont				"-*-Helvetica-Bold-r-*-*-15-*-72-72-p-*-*-*"
#	variable listFont				"-*-swiss 742-*-r-*-*-17-*-85-85-p-*-*-*"
	variable majorSkillColor		PaleTurquoise1
	variable minorSkillColor		gold
	variable popupBackground		#48f500000000
	variable popupForeground		White
	variable primarySkillColor		LightSkyBlue
	variable promptFont				"-*-Helvetica-*-r-*-*-17-*-85-85-p-*-*-*"
	variable rowBackground			ShellBackground
	variable rowForeground			White
	variable scrollBackground		IconForeground
	variable sectionBackground		maroon4
	variable sectionForeground		White
	variable selectBackground		orange3
	variable selectForeground		#48f500000000
	variable statusBackground		#8989a1
	variable statusForeground		White
	variable statusFont				"-*-Helvetica-Bold-R-Normal--*-120-*-*-*-*-*-*"
	variable _textDesc1				""
	variable _textDesc2				""
	variable textFont				"-Misc-Fixed-Bold-R-Normal--18-100-110-*-C-*-ISO8859-1"
	variable value					{}

	variable bold
	variable chunks
	variable currentChunk			{}
	variable dialog
	variable _dump
	variable image1
	variable normal
	variable rawData
	variable _slot
	variable sortedChunks

	namespace export *
}

#	Tune the fonts for the (broken) Windows port.  Use standard
#	control panel colors whenever feasible.

if {"$tcl_platform(platform)" == "windows"} then {
	variable ::Arena2DataHeader::hasExpand	0
	variable ::Arena2Config::hasExtraFont	0

	variable ::ChunkView::statusFont System
	variable ::ChunkView::promptFont System
	variable ::ChunkView::listFont System
	variable ::ChunkView::textFont $::DumpView::_font

	variable ::ChunkView::applicationBackground		SystemAppWorkspace
	variable ::ChunkView::background				SystemWindow
	variable ::ChunkView::buttonActiveBackground	SystemButtonFace
	variable ::ChunkView::buttonActiveForeground	SystemButtonText
	variable ::ChunkView::buttonBackground			SystemButtonFace
	variable ::ChunkView::buttonForeground			SystemButtonText
	variable ::ChunkView::canAnimate				[file executable txttoppm.exe]
	variable ::ChunkView::checkColor				gold
	variable ::ChunkView::columnBackground			SystemBackground
	variable ::ChunkView::columnForeground			gold
	variable ::ChunkView::computeTrough				gold2
	variable ::ChunkView::daggerTrough				SteelBlue
	variable ::ChunkView::editBackground			MidnightBlue
	variable ::ChunkView::editDisabledBackground	SteelBlue
	variable ::ChunkView::editDisabledForeground	White
	variable ::ChunkView::editForeground			White
	variable ::ChunkView::foreground				SystemWindowText
	variable ::ChunkView::legendBackground			SteelBlue
	variable ::ChunkView::majorSkillColor			PaleTurquoise1
	variable ::ChunkView::minorSkillColor			gold
	variable ::ChunkView::popupBackground			SystemMenu
	variable ::ChunkView::popupForeground			SystemMenuText
	variable ::ChunkView::primarySkillColor			LightSkyBlue
	variable ::ChunkView::rowBackground				SystemWindow
	variable ::ChunkView::rowForeground				SystemWindowText
	variable ::ChunkView::scrollBackground			#5470aa
	variable ::ChunkView::sectionBackground			SystemHighlight
	variable ::ChunkView::sectionForeground			SystemHighlightText
	variable ::ChunkView::selectBackground			SystemHighlight
	variable ::ChunkView::selectForeground			SystemHighlightText
} elseif {[catch {winfo rgb . ShellBackground}]} then {
#	Handle stock *nix setups.

	variable ::Arena2Config::hasExtraFont			0
	variable ::ChunkView::extraSpace				2
	variable ChunkView::applicationBackground		#7e88ab
	variable ChunkView::background					#7e88ab
	variable ChunkView::buttonActiveBackground		#7c98d3
	variable ChunkView::buttonBackground			#5470aa
	variable ChunkView::columnBackground			#7e88ab
	variable ChunkView::rowBackground				#7e88ab
	variable ChunkView::scrollBackground			#5470aa
} else {
#	Handle in-house setups.

	variable ::Arena2Config::hasExtraFont	1
	variable ::ChunkView::extraSpace		1
#	variable ::ChunkView::listFont			"-*-Helvetica-Bold-r-*-*-15-*-72-72-p-*-*-*"
	variable ::ChunkView::listFont			"-*-swiss 742-*-r-*-*-17-*-85-85-p-*-*-*"
}

#							 *****************
#							** Class methods **
#							 *****************

proc ChunkView::canInvoke {saveId} {
#		Answer true if no other instance of the receiver
#		is alive.

	return [expr 1 - [winfo exists .cv$saveId]]
}

proc ChunkView::dumpColumn {w fields} {
#		Format the contents of a chunk header on the screen.

	set dump $w.dump.pane
	$dump delete 1.0 end
	foreach line $fields {
		$dump insert end $line
	}
}

proc ChunkView::invoke {saveId {status {}}} {
#		Create and display a fresh instance of the receiver.

	dialog [new] $saveId $status
}

proc ChunkView::new {} {
#		Create a fresh instance of a data space bound to the
#		methods of this class.  Inherit the initial values of
#		the new object from the base configuration.  Asnwer the
#		initialized data space.

	variable _counter
	variable applicationBackground
	variable background
	variable buttonActiveBackground
	variable buttonActiveForeground
	variable buttonBackground
	variable buttonForeground
	variable checkColor
	variable canAnimate
	variable currentChunk
	variable columnBackground
	variable columnForeground
	variable computeTrough
	variable daggerTrough
	variable editBackground
	variable editDisabledBackground
	variable editDisabledForeground
	variable editForeground
	variable extraSpace
	variable field
	variable foreground
	variable gamma
	variable legendBackground
	variable listBackground
	variable listFont
	variable listFont
	variable majorSkillColor
	variable minorSkillColor
	variable popupBackground
	variable popupForeground
	variable primarySkillColor
	variable promptFont
	variable rowBackground
	variable rowForeground
	variable scrollBackground
	variable sectionBackground
	variable sectionForeground
	variable selectBackground
	variable selectForeground
	variable statusBackground
	variable statusForeground
	variable statusFont
	variable _textDesc1
	variable _textDesc2
	variable textFont
	variable tmpDir
	variable value

#	Make a (trivially) unique name for the data space.

	set name ChunkView$_counter
	incr _counter

#	Note the funky double empty lists are required to survive
#	passage through the Tcl parser.

	namespace eval :: namespace eval $name \
		[list variable \
			_textDesc1				\"$_textDesc1\" \
			_textDesc2				\"$_textDesc2\" \
			animes					{{}} \
			applicationBackground	$applicationBackground \
			background				$background \
			buttonActiveBackground	$buttonActiveBackground \
			buttonActiveForeground	$buttonActiveForeground \
			buttonBackground		$buttonBackground \
			buttonForeground		$buttonForeground \
			canAnimate				$canAnimate \
			checkColor				$checkColor \
			columnBackground		$columnBackground \
			columnForeground		$columnForeground \
			computeTrough			$computeTrough \
			currentChunk			{{}} \
			daggerTrough			$daggerTrough \
			editBackground			$editBackground \
			editDisabledBackground	$editDisabledBackground \
			editDisabledForeground	$editDisabledForeground \
			editForeground			$editForeground \
			extraSpace				$extraSpace \
			field					{{}} \
			foreground				$foreground \
			gamma					$gamma \
			legendBackground		$legendBackground \
			listBackground			$listBackground \
			listFont				\"$listFont\" \
			majorSkillColor			$majorSkillColor \
			minorSkillColor			$minorSkillColor \
			popupBackground			$popupBackground \
			popupForeground			$popupForeground \
			primarySkillColor		$primarySkillColor \
			promptFont				\"$promptFont\" \
			rowBackground			$rowBackground \
			rowForeground			$rowForeground \
			scrollBackground		$scrollBackground \
			sectionBackground		$sectionBackground \
			sectionForeground		$sectionForeground \
			selectBackground		$selectBackground \
			selectForeground		$selectForeground \
			self					::$name \
			statusBackground		$statusBackground \
			statusFont				$statusFont \
			statusForeground		$statusForeground \
			textFont				\"$textFont\" \
			tmpDir					$tmpDir \
			value					{{}} \
		]
	set name ::$name
	namespace eval $name {
		namespace import ::ChunkView::*
	}
	return $name
}

proc ChunkView::picToTxt { pic } {
#		Convert the given picture code into a texture file
#		resource number and entry within that resource pool.

	set resNo [expr $pic >> 7]
	if { $resNo < 0 } then {
		set delta [expr 256 + $resNo]
		set resNo [expr 256 + $delta]
	}
	lappend reply $resNo
	lappend reply [expr $pic & 127]
	return $reply
}

proc ChunkView::positionWindow w {
#		Position the viewer window at a fixed position on the screen.

	wm geometry $w +300+300
}

proc ChunkView::searchFailed {} {
#		Stock search failed message box.

	tk_messageBox \
		-icon info \
		-type ok \
		-title "Search failed" \
		-message "Requested link not found."
}

proc ChunkView::showStatus {w column index} {
#		Show the name of the program in the status bar. This procedure
#		is called when the user moves the cursor over a browser description.

	set t $w.$column.dump.link
	set tags [$t tag names $index]
	set i [lsearch -glob $tags chunks-*]
	variable ::Arena2DataHeader::links
	if {$i < 0} {
		$w.statusBar.lab config -text " "
	} else {
		set viewer [string range [lindex $tags $i] 7 end]
		set statusHelp [helpText $viewer]
		if {[keylget ::Arena2DataHeader::links $viewer link]} then {
			$w.statusBar.lab config -text "$statusHelp (Jump to)"
		} else {
			if { [lsearch -exact $tags chunks] < 0 } {
				$w.statusBar.lab config -text "$statusHelp"
			} else {
				$w.statusBar.lab config -text "$statusHelp (Sort by)"
			}
		}
	}
}

#							 ********************
#							** Instance methods **
#							 ********************

proc ChunkView::animators {self} {
#		Answer a list of all the running animations
#		in the receiver.

	set reply {}
	set _slot [namespace eval $self concat \$_slot]
	if {[llength [namespace which -variable $self\::a$_slot]] > 0} then {
		eval lappend reply $$self\::a$_slot
	}
	if {[llength [namespace which -variable $self\::b$_slot]] > 0} then {
		eval lappend reply $$self\::b$_slot
	}
	return "$reply"
}

proc ChunkView::dialog {self saveId {status {}}} {
#		Build the main display dialog.

	set title ".cv$saveId \"Chunk Viewer $status\""
	namespace eval $self {

#		Leave only one instance of the app alive in the same process.

		variable _slot [uplevel concat \$saveId]
		variable w .cv$_slot
		catch {destroy $w}
		toplevel $w \
				-background $applicationBackground \
				-class ChunkView
		uplevel eval wm title \$title
		positionWindow $w
		variable dialog $w
#		frame	$dialog \
#				-background $applicationBackground
#		pack	$dialog \
#				-side top \
#				-expand yes \
#				-fill both

#		Add a status bar to display hot link information and a frame
#		to hold the pick list and chunk data display panes.

		variable w $dialog.chunk
		frame	$dialog.statusBar \
				-background $statusBackground
		frame	$w \
				-background $applicationBackground
		frame	$dialog.f \
				-background $applicationBackground
		pack	$dialog.statusBar \
				-side bottom \
				-fill x \
				-pady 2
		pack	$w $dialog.f \
				-side left \
				-anchor s \
				-expand 0 \
				-fill y

#		Partition the data display into four areas: three columns and a
#		covering area for the hex dump data.  Use a fixed pitch font for
#		the hex dump pane.  The first 2 columns will hold the field name
#		hot links for the chunk header, and the third will display
#		any texture animations bound to the chunk.

		frame	$dialog.f.col1
		frame	$dialog.f.col2
		frame	$dialog.f.col3 \
				-background $applicationBackground
		frame	$dialog.f.data
		pack	$dialog.f.data \
				-side top \
				-anchor nw \
				-expand 1 \
				-fill y
		pack	$dialog.f.col1 $dialog.f.col2 \
				-side left \
				-anchor nw \
				-expand 0 \
				-fill none
		pack	$dialog.f.col3 \
				-side left \
				-anchor nw \
				-expand 1 \
				-fill both
	
		label	$dialog.statusBar.lab \
				-text "   " \
				-relief sunken \
				-bd 2 \
				-font $statusFont \
				-anchor w \
				-background $statusBackground \
				-foreground $statusForeground
		label	$dialog.statusBar.sort \
				-text "Sort by offset" \
				-width 20 \
				-relief sunken \
				-bd 2 \
				-font $statusFont \
				-anchor w \
				-background $statusBackground \
				-foreground $statusForeground
		pack	$dialog.statusBar.lab \
				-side left \
				-padx 2 \
				-expand yes \
				-fill both
		pack	$dialog.statusBar.sort \
				-side left \
				-padx 2

#		Build the picklist panes for the chunk records.  First a caption.
#		Use a larger font to stand out.

		label	$w.msg \
				-background $applicationBackground \
				-font {Helvetica 14} \
				-wraplength 4i \
				-justify left \
				-text "Select a chunk"
	    pack	$w.msg \
	    		-side top

#		Exit button.

		frame	$w.buttons \
				-background $applicationBackground
		pack	$w.buttons \
				-side bottom \
				-fill x \
				-pady 2m \
				-padx 2m
		button	$w.buttons.dismiss \
				-background $background \
				-text Close \
				-command "cvCloseDialog $self $dialog" \
				-font $listFont
		pack	$w.buttons.dismiss \
				-side left -expand 1

#		List box header caption.

		frame	$w.spacer1 \
				-background $applicationBackground \
				-height 3m \
				-width 20
		label	$w.chunkLabel \
				-background $applicationBackground \
				-text "Object Id       Description" \
				-font $listFont
		pack	$w.spacer1 $w.chunkLabel \
	    		-side top \
	    		-anchor w
		frame	$w.f
		pack	$w.f \
				-side top \
				-expand yes \
				-fill both

#		The list box and its scrollbar control

		listbox	$w.f.list \
				-relief sunken \
				-width 30 \
				-height 20 \
				-selectmode browse \
				-selectbackground $selectBackground \
				-selectforeground $selectForeground \
				-exportselection false \
				-yscrollcommand "$w.f.scroll set" \
				-font $listFont
		scrollbar $w.f.scroll \
				-troughcolor $scrollBackground \
				-background $buttonActiveBackground \
				-activebackground $buttonActiveForeground \
				-command "$w.f.list yview"
		pack	$w.f.scroll \
				-side right \
				-fill y
		pack	$w.f.list \
				-side left \
				-fill both \
				-expand 1

#		Bind the usual mouse button to display the list box selection
#		in the dump pane.  Augment the stock listbox behavior so the
#		keyboard traversal of the picklist will display the chunks
#		as well.

		bind $w.f.list	<Button-1>	"+ChunkView::showChunk $self @%x,%y"
		bind Listbox	<Up>		"+ChunkView::showChunk $self active"
		bind Listbox	<Down>		"+ChunkView::showChunk $self active"
#		bind $w.f.list	<Double-1>	"+showFaction $w %x %y"
#		bind $w.f.list	<Button-2>	"+toggleHiddenFactions $w %x %y"

#		Add a search menu to the list box and bind button 3 to pop
#		it up over the list box.

		variable _menu [uplevel concat .cvmenu\$saveId]
		menu	$_menu \
				-background $popupBackground \
				-foreground $popupForeground \
				-activebackground $selectBackground \
				-activeforeground $selectForeground \
				-disabledforeground $popupForeground \
				-relief raised \
				-bd 3 \
				-tearoff 0
		menu	$_menu.keys \
				-background $popupBackground \
				-foreground $popupForeground \
				-activebackground $selectBackground \
				-activeforeground $selectForeground \
				-disabledforeground $popupForeground \
				-relief raised \
				-bd 3 \
				-tearoff 0
		}
#		Prune empty chunk name entries for the cascading search menu.

		set index 0
		set keys {}
		foreach key $::Arena2DataHeader::names {
			if {[clength $key] > 0} then {
				set ind [format %x $index]
				lappend keys \
						"-label \"$key\" \
						-command \"ChunkView::searchFor $self key 0x$ind 1\" \
						-foreground White "
			}
			incr index
		}
#		Sort the chunk key names alphabetically and build the cascading
#		menu for search items.

		set keys [lsort $keys]
		set index 0
		eval set _menu $$self\::_menu
		foreach key $keys {
			eval [concat $_menu.keys add command \
				$key \
				-columnbreak [expr ($index & 7) == 0 && $index > 0]]
			incr index
		}
		namespace eval $self {
		$_menu	add command \
				-command {} \
				-label "Search for" \
				-state disabled
		$_menu	add separator
		$_menu	add command \
				-command "ChunkView::findOid $self" \
				-label "Object ID" \
				-foreground White \
				-underline 0
		$_menu	add cascade \
				-label "Description" \
				-foreground $popupForeground \
				-underline 0 \
				-menu $_menu.keys
#		$_menu	add command \
#				-command "ChunkView::findKey" \
#				-label "Description" \
#				-foreground $popupForeground \
#				-underline 0
		$_menu	add separator
		$_menu	add command \
				-command "ChunkView::nextKey $self" \
				-label "Next" \
				-foreground $popupForeground \
				-underline 0
		$_menu	add command \
				-command "ChunkView::priorKey $self" \
				-label "Previous" \
				-foreground $popupForeground \
				-underline 0
		bind $w.f.list <Button-3> "+tk_popup $_menu %X %Y"

#		Add a pane for the hot links and the field values.

		frame	$dialog.f.col1.dump
		frame	$dialog.f.col2.dump
		pack	$dialog.f.col1.dump \
				-side left \
				-fill both \
				-expand 1
		pack	$dialog.f.col2.dump \
				-side left \
				-fill both \
				-expand 1
	}
#	Add the hot links for the field names.

	eval dialogTags $self f.col1 $$self\::w
	eval dialogTags $self f.col2 $$self\::w

#	Bind the mouse motion to the status bar display.

	.cv$saveId.f.col1.dump.link tag bind status <Enter> \
		"cvEnter $self .cv$saveId f.col1 f.col1.dump.link @%x,%y"
	.cv$saveId.f.col1.dump.link tag bind status <Motion> \
		"cvMotion $self .cv$saveId f.col1 f.col1.dump.link @%x,%y"
	.cv$saveId.f.col2.dump.link tag bind status <Enter> \
		"cvEnter $self .cv$saveId f.col2 f.col2.dump.link @%x,%y"
	.cv$saveId.f.col2.dump.link tag bind status <Motion> \
		"cvMotion $self .cv$saveId f.col2 f.col2.dump.link @%x,%y"

#	Image panes for the item pictures.  Add status bar bindings to show the
#	texture file description.

	namespace eval $self {
		catch {image delete a$_slot}
		image create photo a$_slot -gamma $gamma
		catch {image delete b$_slot}
		image create photo b$_slot -gamma $gamma

		variable image1 $dialog.f.col3
		frame	$image1.bottom \
				-background $applicationBackground
		pack	$image1.bottom \
				-side bottom \
				-expand yes \
				-fill both \
				-padx 2m
		label	$image1.image1Label \
				-background $applicationBackground \
				-text "  3D view" \
				-justify left \
				-font $listFont
		label	$image1.itemLabel \
				-background $applicationBackground \
				-text "Inventory  " \
				-justify left \
				-font $listFont 
		pack	$image1.image1Label \
				-side left \
				-pady 2 \
				-anchor n
		pack	$image1.itemLabel	\
				-side right \
				-pady 2 \
				-anchor n

		label	$image1.bottom.image1 \
				-background $applicationBackground \
				-image a$_slot \
				-bd 2
		label	$image1.bottom.item \
				-background $applicationBackground \
				-image b$_slot \
				-bd 2
		pack	$image1.bottom.image1 \
				-side left \
				-pady 2 \
				-padx 2 \
				-ipadx 2 \
				-ipady 2
		pack	$image1.bottom.item \
				-side right \
				-pady 2 \
				-padx 2 \
				-ipadx 2 \
				-ipady 2

		bind $image1.bottom.image1	<Enter>	"::ChunkView::textureStatus $self $dialog textDesc1"
		bind $image1.bottom.image1	<Leave>	"$dialog.statusBar.lab config -text \" \""
		bind $image1.bottom.item	<Enter>	"::ChunkView::textureStatus $self $dialog textDesc2"
		bind $image1.bottom.item	<Leave>	"$dialog.statusBar.lab config -text \" \""
	}
#	Add the hex dump pane.

	set _dump [DumpView::openInside .cv$saveId.f.data]
	namespace eval $self variable _dump $_dump

#	Modal lock until the dialog is populated.

	busy hold .cv$saveId
	update
	loadChunks $self .cv$saveId.chunk.f.list $saveId
	initLinks $self .cv$saveId
	busy release .cv$saveId
	update
}

proc ChunkView::dialogTags {self column w} {
#		Add the dialog tags/bindings for a pane set.

	namespace eval $self variable \
		valuePane \$dialog.$column.dump.pane \
		linkPane \$dialog.$column.dump.link \
		t \$dialog.$column.dump.link
	namespace eval $self {
		text	$valuePane \
				-background $applicationBackground \
				-width 10 \
				-height 11 \
				-spacing3 1 \
				-font $textFont
		text	$linkPane \
				-background $applicationBackground \
				-width 13 \
				-height 12 \
				-setgrid true \
				-wrap none \
				-cursor [$w cget -cursor] \
				-font $listFont
		pack	$linkPane $valuePane \
				-side left \
				-anchor n \
				-fill y

#		Create tags for the hot link buttons.

		$t tag configure chunksspace -justify right
		if {[winfo depth .] == 1} {
			$t tag configure chunks \
				-lmargin1 1c \
				-lmargin2 1c \
				-justify right \
				-underline 1
			$t tag configure visited \
				-lmargin1 1c \
				-lmargin2 1c \
				-underline 1
			variable bold \
				"-background black \
				-foreground white"
			variable normal
				"-background {} \
				-foreground {}"
		} else {
			$t tag configure chunks \
				-lmargin1 1c \
				-lmargin2 1c \
				-foreground $buttonActiveForeground \
				-background $buttonBackground \
				-justify right \
				-relief sunken \
				-offset $extraSpace \
				-borderwidth 2
			$t tag configure visited \
				-lmargin1 1c \
				-lmargin2 1c \
				-foreground $buttonActiveForeground \
				-background $buttonBackground \
				-relief sunken \
				-offset $extraSpace \
				-borderwidth 2
			variable bold \
				"-foreground $sectionForeground \
				-background $sectionBackground \
				-relief raised \
				-borderwidth 2"
			variable normal  \
				"-background {} \
				-foreground {} \
				-relief sunken"
		}
		$t tag configure inert \
			-justify right \
			-offset $extraSpace \
			-lmargin1 1c \
			-lmargin2 1c

#		Add a binding to track the changing status messages.

		$t tag bind status <Leave> \
			"$dialog.statusBar.lab config -text \"\" "
	}
	global lastLine
	set lastLine ""
}

proc ChunkView::discard {self} {
#		Permanently remove an instance and its references.

#	Clear the debris from the animator.  The file cmd is dim-witted,
#	so avoid giving it an empty list to process.

	set blob [namespace eval $self concat \
		[file join \$tmpDir "pic"] \
		[file join \$tmpDir "pic.ppm"] \
		\[glob -nocomplain \[file join \$tmpDir \"\${_slot}p*.*\"\]\]]
	if {[llength $blob] > 0} then {
		eval file delete $blob
	}
	namespace delete $self
}

proc ChunkView::dumpChunk {self w} {
#		Format the contents of a chunk header on the screen.

	clearAnimation [animators $self]
	eval set canAnimate $$self\::canAnimate
	eval set currentChunk $$self\::currentChunk
	eval set dialog $$self\::dialog
	eval set _dump $$self\::_dump
	eval set image1 $$self\::image1
	eval set rawData $$self\::rawData
	eval set _slot $$self\::_slot
	eval set tmpDir $$self\::tmpDir

#	Populate the header field values from the current chunk.
#	Dump the remaining data bytes in the chunk in the hex pane.

	set fields [printValues $currentChunk]
	set half [expr ([llength $fields] + 1) / 2]
	dumpColumn $w.f.col1 [lrange $fields 0 [expr $half - 1]]
	dumpColumn $w.f.col2 [lrange $fields $half end]
	contents: $_dump [keylget rawData [keylget currentChunk "offset"]]

#	Display any animations attached to this chunk.  If the chunk is
#	an item chunk, grab the item animators from the item data block
#	rather than the chunk header.

	if {[keylget currentChunk "key"] == 2} then {
#		Use the item detailed inventory picture instead of the header entries.

		set bytes [lrange [rawContents $_dump] 50 53]
		set pic2 [expr 0x[lindex $bytes 0] + (256 * 0x[lindex $bytes 1])]
		set pic1 [expr 0x[lindex $bytes 2] + (256 * 0x[lindex $bytes 3])]
	} else {
		set pic1 [keylget currentChunk "picture1"]
		set pic2 [keylget currentChunk "picture2"]
	}
#	Some chunks reuse the picture fields for other business.  Avoid
#	the obviously bogus picture values, but note that txtImageAt
#	will replace any bogus entry it finds with a stock image.
#	Special handling in case the texture accellerator is unavailable.

	if { $canAnimate && ($pic1 > 0x300 || $pic1 < -2) } then {
		set labl [txtImageAt $self \
							 [picToTxt $pic1] \
							 [file join "$tmpDir" "${_slot}p1.00"] \
							 textDesc1:]
		$image1.image1Label configure -text "Texture [lindex $labl 0]"
		$image1.bottom.image1 configure -relief sunken
		animate $self a$_slot [lindex $labl 1]
	} else {
		$image1.image1Label configure -text "  3D view"
		set $self\::_textDesc1 ""
		$image1.bottom.image1 configure -relief flat
		a$_slot blank
	}
	if { $canAnimate && ($pic2 > 0x300 || $pic2 < -2) } then {
		set labl [txtImageAt $self \
							 [picToTxt $pic2] \
							 [file join "$tmpDir" "${_slot}p2.00"] \
							 textDesc2:]
		$image1.itemLabel configure -text "Texture [lindex $labl 0]"
		if {[llength [lindex $labl 1]] > 1 && $pic1 == $pic2 } then {
			sleep 1
		}
		$image1.bottom.item configure -relief sunken
		animate $self b$_slot [lindex $labl 1]
	} else {
		$image1.itemLabel configure -text "Inventory  "
		set $self\::_textDesc2 ""
		$image1.bottom.item configure -relief flat
		b$_slot blank
	}
}

proc ChunkView::findKey {self} {
#		Prompt and search for a chunk key.

	eval set promptFont $$self\::promptFont
	set reply [tk_prompter .prompter "Chunk Search" "Enter key/description" $promptFont {}]
	if {[clength "$reply"] < 1} {
		return
	}
	if {[crange "$reply" 0 1] == "0x"} then {
#		Do nothing.
		set key {}
	} elseif {[ctype digit "$reply"]} then {
#		Do nothing.
		set key {}
	} elseif {[ctype xdigit "$reply"]} then {
		set reply 0x$reply
	} else {
		set key [name2Key "$reply"]
		if {$key < 0} {
			tk_messageBox \
				-icon info \
				-type ok \
				-title "Search failed" \
				-message "Description \"$reply\" not found."
			return
		}
		set reply $key
	}
	searchFor $self key $reply 1
}

proc ChunkView::findOid {self} {
#		Prompt and search for an objectId.

	eval set promptFont "$$self\::promptFont"
	set reply [tk_prompter .prompter "Chunk Search" "Enter object ID" $promptFont  {}]
	set len [clength $reply]
	if {[clength $reply] < 1} {
		return
	}
	if {[clength $reply] != 8} {
		tk_messageBox \
			-icon info \
			-type ok \
			-title "Search failed" \
			-message "Object ID not found."
		return
	}
	searchFor $self objectId 0x$reply 1
}

proc ChunkView::initColumn {self w dialog} {
#		Initialize the hot links.

	eval set bold $$self\::bold
	eval set normal $$self\::normal
	set link $w.dump.link
	$link insert end " " {chunksspace}
	set invokers {{link {jumpTo hand1}} {sort {sortBy rtl_logo}}}
	foreach segment $dialog {
		set widget [lindex $segment 0]
		set text [lindex $segment 1]
		set canInvoke [keylget invokers $widget invoke]
		if {!$canInvoke || "$widget" == "text" } then {
			$link insert end ${text}: [list inert chunks-$text status]
		} else {
			set cursor [lindex $invoke 1]
			set invoke [lindex $invoke 0]
			$link insert end " $text " [list chunks chunks-$text status]
			$link tag bind chunks-$text <Any-Enter> \
				"$link tag configure chunks-$text $bold ; \
				 $link config -cursor $cursor"
			$link tag bind chunks-$text <Any-Leave> \
				"$link tag configure chunks-$text $normal ; \
				 $link config -cursor {}"
			$link tag bind chunks-$text <1> \
				"ChunkView::$invoke $self $text"
		}
		$link insert end "\n " {chunksspace}
	}
	$link configure -state disabled
}

proc ChunkView::initLinks {self w} {
#		Initialize the hot links.

	set dialog [Arena2DataHeader::widgetList]
	set half [expr ([llength $dialog] + 1) / 2]
	initColumn $self $w.f.col1 [lrange $dialog 0 [expr $half - 1]]
	initColumn $self $w.f.col2 [lrange $dialog $half end]
}

proc ChunkView::jumpTo {self keyName} {
#		Display the chunk referenced by a key.

	eval set currentChunk $$self\::currentChunk
	eval set sortedChunks $$self\::sortedChunks
	eval set dialog $$self\::dialog
	variable ::Arena2DataHeader::links
	set pane $dialog.chunk.f.list
	keylget currentChunk $keyName key
	keylget ::Arena2DataHeader::links $keyName kName
	set index 0
	foreach chunk $sortedChunks {
		keylget chunk $kName aKey
		if {$aKey == $key} then {
			set curSel [lindex [$pane curselection] 0]
			set $self\::currentChunk $chunk
			$pane selection clear $curSel
			$pane see $index
			$pane activate $index
			$pane selection set $index
			dumpChunk $self $dialog
			update
			return
		}
		incr index
	}
	searchFailed
}

proc ChunkView::loadChunks {self pane saveId} {
#		Fill the contents of the list pane with the chunk names
#		in the receiver.

	set theWorks [chunksFor $saveId]
	namespace eval $self {
		variable chunks [uplevel lindex \$theWorks 0]
		variable sortedChunks [uplevel lindex \$theWorks 0]
		variable rawData [uplevel lindex \$theWorks 1]
	}
	$pane delete 0 end

#	Use the default sort order by offset.

	set sortedChunks [lindex $theWorks 0]
#	set sortedChunks [lsort -command {sortChunk} $chunks]
	foreach chunk $sortedChunks {
		$pane insert end [Arena2DataHeader::listString $chunk]
	}
}

proc ChunkView::nextKey {self} {
#		Scroll the list box selection to the next matching entry.

	eval set field $$self\::field
	eval set value $$self\::value
	eval set currentChunk $$self\::currentChunk
	if {[clength "$field"] < 1} {
		if {[llength $currentChunk] > 0} then {
			set field key
			keylget currentChunk "key" value
		} else {
			tk_messageBox -icon info \
				-type ok \
				-title "Search failed" \
				-message "There is no previous match."
			return
		}
	}
	searchFor $self $field $value 1
}

proc ChunkView::priorKey {self} {
#		Scroll the list box selection to the previous matching entry.

	eval set field $$self\::field
	eval set value $$self\::value
	eval set currentChunk $$self\::currentChunk
	if {[clength "$field"] < 1} {
		if {[llength $currentChunk] > 0} then {
			set field key
			keylget currentChunk "key" value
		} else {
			tk_messageBox -icon info \
				-type ok \
				-title "Search failed" \
				-message "There is no previous match."
			return
		}
	}
	searchFor $self $field $value -1
}

proc ChunkView::searchFor {self newField newValue direction} {
#		Search the sorted chunks for the next occurance of the
#		specified field value.  If direction is +1, search
#		forwards from the current selection.  If direction is -1,
#		search backwards from the current selection.

	eval set dialog $$self\::dialog
	busy hold $dialog
	update
	eval set field $$self\::field
	eval set value $$self\::value
	eval set sortedChunks $$self\::sortedChunks
	eval set currentChunk $$self\::currentChunk
	set pane $dialog.chunk.f.list
	set sel0 [$pane curselection]
	set sel [expr $sel0 + $direction]
	set length [llength $sortedChunks]
	if {$sel < 0 || $sel >= $length} {
		busy release $dialog
		update
		tk_messageBox \
			-icon info \
			-type ok \
			-title "Search failed" \
			-message "There are no more matchs."
		return
	}
	while {$length > 0} {
		set chunk [lindex $sortedChunks $sel]
		if {[llength $chunk] == 0} then {
			break
		}
		if {[keylget chunk $newField] == $newValue} {
			if {[clength "$sel0"] > 0} then {
				$pane selection clear $sel0
			}
			$pane see $sel
			$pane activate $sel
			$pane selection set $sel
			namespace eval $self variable \
				currentChunk [list $chunk] \
				field $newField \
				value $newValue
			dumpChunk $self $dialog
			busy release $dialog
			update
			return
		}
		incr sel $direction
		incr length -1
	}
	busy release $dialog
	update
	tk_messageBox \
		-icon info \
		-type ok \
		-title "Search failed" \
		-message "There is no match."
}

proc ChunkView::showChunk {self slot} {
#		Display the header under the selection in the
#		header chunk pane.

	namespace eval $self {
		busy hold $dialog
		update
		variable currentChunk \
			[lindex $sortedChunks [$dialog.chunk.f.list index [uplevel concat \$slot]]]
		dumpChunk $self $dialog
		busy release $dialog
		update
		focus -force $dialog.chunk.f.list
	}
}

proc ChunkView::sortBy {self keyName} {
#		Sort the saved game chunks by a keyName and repopulate
#		the picklist with the new order.  Modal lock the application
#		until the sort is finished.

	namespace eval $self {
		variable keyName [uplevel concat \$keyName]
		$dialog.statusBar.sort configure -text "Sort by $keyName"
		busy hold $dialog
		update
		$dialog.chunk.f.list delete 0 end
		variable sortedChunks [lsort -command [list sortByKey $keyName] $chunks]
		foreach chunk $sortedChunks {
			$dialog.chunk.f.list insert end [Arena2DataHeader::listString $chunk]
		}
		namespace eval $_dump {
			$_dialog.pane delete 1.0 end
			$_dialog.pane delete 1.0 end
			$_dialog.pane delete 1.0 end
		}
		clearAnimation [animators $self]
		$image1.itemLabel configure -text "Inventory  "
		$image1.image1Label configure -text "  3D view"
		a$_slot blank
		b$_slot blank
		textDesc1: $self ""
		textDesc2: $self ""
		busy release $dialog
		update
	}
}

proc ChunkView::textDesc1 {self} {
#		Answer the first image description.

	namespace eval $self { return $_textDesc1}
}

proc ChunkView::textDesc1: {self arg} {
#		Assign the first image description.

	namespace eval $self variable _textDesc1 $arg
}

proc ChunkView::textDesc2 {self} {
#		Answer the second image description.

	namespace eval $self { return $_textDesc2}
}

proc ChunkView::textDesc2: {self arg} {
#		Assign the second image description.

	namespace eval $self variable _textDesc2 $arg
}

proc ChunkView::textureStatus {self w arg} {
#		Show the name of the texture file in the status bar. This procedure
#		is called when the user moves the cursor over a texture image pane.

	set text [namespace eval $self $arg $self]
	if {[clength $text] < 1} {
		$w.statusBar.lab config -text " "
	} else {
		$w.statusBar.lab config -text "$text"
	}
}

proc ChunkView::txtImageAt {self resource dest status} {
#		Retrieve the specified texture flipbook from
#		the Daggerfall installation.

	set resNo [lindex $resource 0]
	set serialNo [lindex $resource 1]
	set texture [Arena2Texture::buildTables $resNo]
	set res $resource
	if {[llength $texture] == 0} then {
		set bad 1
		set count 0
	} elseif {[llength [lindex $texture 0]] < $serialNo} then {
		set bad 1
		set count [llength [lindex $texture 1]]
	} else {
		set bad 0
	}
	if {!$bad} then {
		set image [imagesAt $texture $serialNo $dest]
		if {[llength $image] < 1} then {
			set bad 1
		}
	}
	if {$bad} then {
#		If the requested sequence doesn't exist, give him
#		a Daedra prince by default.

#		puts "texture.$resNo:$serialNo vs header count $count is wrong."
		set resNo 175
		set serialNo 15
		set res 0x578f
		set texture [Arena2Texture::buildTables $resNo]
		set image [imagesAt $texture $serialNo $dest]
	}
	if {[Arena2Texture::currentRate] < 80} then {
		if { $resNo > 380 && $resNo < 400} then {
			currentRate: [expr 2000 - [currentRate]]
		}
		if { $resNo > 450 && $resNo < 457} then {
			currentRate: [expr 2000 - [currentRate]]
		}
	}
	lappend reply "$resNo:$serialNo"
	lappend reply $image
	lappend reply "$res"
#	lappend reply [currentRate]
	$status $self "[lindex $texture 2]"
	return $reply
}

#						 **********************
#						** Global dispatchers **
#						 **********************

proc cvCloseDialog { self w } {
#		Destroy the view of a saved game.

#	Clear the debris from the animator.

	Arena2Texture::clearAnimation [ChunkView::animators $self]

#	Finally, remove the ensemble of widgets from the screen.

	destroy $w
	namespace eval $self {
		destroy $_menu
		DumpView::discard $_dump
	}
	ChunkView::discard $self
}

proc cvEnter {self dialog frame pane selPos} {
#		Capture the status pane display on entry to a hot-spot.

	global lastLine
	set lastLine [$dialog.$pane index [concat $selPos linestart]]
	eval ChunkView::showStatus $$self\::dialog $frame [$dialog.$pane index $selPos]
}

proc cvMotion {self dialog frame pane selPos} {
#		Capture the status pane change on movement through a hot-spot.

	global lastLine
	global newLine

	set newLine [$dialog.$pane index [concat $selPos linestart]]
	if {[string compare $newLine $lastLine] != 0} {
		set lastLine $newLine
	}
	eval ChunkView::showStatus $$self\::dialog $frame [$dialog.$pane index $selPos]
}

proc sortByKey { key a b } {
#		Sort two chunk fields numerically.  Note the
#		numerics are actually unsigned, despite the fact
#		they are rendered by signed quantities in Tcl, so
#		partition the phase signature accordingly.

	keylget a $key na
	keylget b $key nb
	if {$na < 0 && $nb < 0} {
		return [expr $na - $nb]}
	if {$na < 0} {return 1}
	if {$nb < 0} {return -1}
	return [expr $na - $nb]
}

proc sortChunk {a b} {
#		Sort two chunk fields by name.

	keylget a "name" na
	keylget b "name" nb
	return [string compare "$na" "$nb"]
}

#							 ******************
#							** Initialization **
#							 ******************

#	Remove the silly default wish window.

eval destroy [winfo child .]
wm title . "TES2 Saved Games"

#	Inherit the installation configuration from the Arena2Config
#	namespace when it is available.

if {[llength [namespace children :: ::Arena2Config]] > 0} then {
	variable Arena2Config::gamma
	variable Arena2Config::tmpDir
	variable ChunkView::gamma "$Arena2Config::gamma"
	variable ChunkView::tmpDir "$Arena2Config::tmpDir"
}

#	Tune the configuration setup from the command line and
#	open the main dialog

if {$clicount > 0} then {
	variable Arena2SaveTree::exeDir
	variable Arena2SaveTree::artDir
	set Arena2SaveTree::exeDir [lindex $argv 0]
	set Arena2SaveTree::artDir [lindex $argv 0]/arena2
}
#	Initialize the stock palette for BS pseudocolor images
#	and open the saved game picklist.

Arena2Image::buildPalette
Arena2SaveGame::saveGameDialog ChunkView
