# prompter.tcl --
#
# This file defines the procedure tk_prompter, which creates a dialog
# box containing a bitmap, a message, an entry field, and an OK button.
#
# RCS: @(#) $Id: dialog.tcl,v 1.3 1998/09/14 18:23:22 stanton Exp $
#
# Copyright (c) 1992-1993 The Regents of the University of California.
# Copyright (c) 1994-1997 Sun Microsystems, Inc.
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

#
# tk_prompter:
#
# This procedure displays a dialog box, waits for the OK button in the dialog
# to be invoked, then returns the entered text to the caller.  If the
# dialog somehow gets destroyed, nothing is returned.
#
# Arguments:
# w -		Window to use for dialog top-level.
# title -	Title to display in dialog's decorative frame.
# text -	Prompt to display in dialog.
# font -	Display font to use for the prompt.
# bitmap -	Bitmap to display in dialog (empty string means none).

proc tk_prompter {w title text font bitmap } {
    global tkPriv tcl_platform

#	1. Create the top-level window and divide it into top
#	and bottom parts.

	catch {destroy $w}
	toplevel $w -class Dialog
	wm title $w $title
	wm iconname $w Dialog
	wm protocol $w WM_DELETE_WINDOW { }

#	The following command means that the dialog won't be posted if
#	[winfo parent $w] is iconified, but it's really needed;  otherwise
#	the dialog can become obscured by other windows in the application,
#	even though its grab keeps the rest of the application from being used.

	wm transient $w [winfo toplevel [winfo parent $w]]
	if {$tcl_platform(platform) == "macintosh"} {
		unsupported1 style $w dBoxProc
	}

	frame $w.bot
	frame $w.top
	if {$tcl_platform(platform) == "unix"} {
		$w.bot configure -relief raised -bd 1
		$w.top configure -relief raised -bd 1
	}
	pack $w.bot -side bottom -fill both
	pack $w.top -side top -fill both -expand 1

#	2. Fill the top part with bitmap and message (use the option
#	database for -wraplength so that it can be overridden by
#	the caller).

	option add *Dialog.msg.wrapLength 3i widgetDefault
	entry $w.entry -relief sunken
	pack $w.entry -in $w.top -side bottom -expand 1 -fill x
	label $w.msg -justify left -text $text
	if {$tcl_platform(platform) == "macintosh"} {
		$w.msg configure -font system
	} else {
		$w.msg configure -font "$font"
	}
	pack $w.msg -in $w.top -side right -expand 1 -fill both -padx 3m -pady 3m
	if {$bitmap != ""} {
		if {($tcl_platform(platform) == "macintosh") && ($bitmap == "error")} {
			set bitmap "stop"
		}
		label $w.bitmap -bitmap $bitmap
		pack $w.bitmap -in $w.top -side left -padx 3m -pady 3m
	}

#	3. Create a row of buttons at the bottom of the dialog.

	set i 0
	foreach but {Ok} {
		button $w.button$i -text $but -command "set tkPriv(button) $i"
		if {$i == 0} {
			$w.button$i configure -default active
		} else {
			$w.button$i configure -default normal
		}
		grid $w.button$i -in $w.bot -column $i -row 0 -sticky ew -padx 10
		grid columnconfigure $w.bot $i
#		We boost the size of some Mac buttons for l&f
		if {$tcl_platform(platform) == "macintosh"} {
			set tmp [string tolower $but]
			if {($tmp == "ok") || ($tmp == "cancel")} {
				grid columnconfigure $w.bot $i -minsize [expr 59 + 20]
			}
		}
		incr i
	}

#	4. Create a binding for <Return> on the dialog if there is a
#	default button.

	bind $w <Return> "
		$w.button0 configure -state active -relief sunken
		update idletasks
		after 100
		set tkPriv(button) 0
	"

#	5. Create a <Destroy> binding for the window that sets the
#	button variable to -1;  this is needed in case something happens
#	that destroys the window, such as its parent window being destroyed.

	bind $w <Destroy> {set tkPriv(button) -1}

#	6. Withdraw the window, then update all the geometry information
#	so we know how big it wants to be, then center the window in the
#	display and de-iconify it.

	wm withdraw $w
	update idletasks
	set x [expr {[winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
		- [winfo vrootx [winfo parent $w]]}]
	set y [expr {[winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
		- [winfo vrooty [winfo parent $w]]}]
	wm geom $w +$x+$y
	wm deiconify $w

#	7. Set a grab and claim the focus too.

	set oldFocus [focus]
	set oldGrab [grab current $w]
	if {$oldGrab != ""} {
		set grabStatus [grab status $oldGrab]
	}
	grab $w
	focus $w.entry
#	focus $w.button0

#	8. Wait for the user to respond, then restore the focus and
#	return the index of the selected button.  Restore the focus
#	before deleting the window, since otherwise the window manager
#	may take the focus away so we can't redirect it.  Finally,
#	restore any grab that was in effect.

	tkwait variable tkPriv(button)
	set reply [$w.entry get]
	catch {focus $oldFocus}
	catch {
#		It's possible that the window has already been destroyed,
#		hence this "catch".  Delete the Destroy handler so that
#		tkPriv(button) doesn't get reset by it.

		bind $w <Destroy> {}
		destroy $w
	}
	if {$oldGrab != ""} {
		if {$grabStatus == "global"} {
			grab -global $oldGrab
		} else {
			grab $oldGrab
		}
	}
	return "$reply"
}
