#!/bin/sh
# \
exec wishx "$0" "$@"

#								 ***********
#								** Summary **
#								 ***********

#					  Copyright 2002, Nagual Enterprises
#				Please see COPYING for the complete GNU license

#	(1)	Title:
#				Daggerfall Saved Game Picklist

#	(2)	Purpose:
#			To present a list of Daggerfall saved games,
#			identified by saved image snapshot and name.
#			And arrange to execute an arbitrary action when
#			a saved gane is selected from the picklist.

#	(3)	Platform dependencies:
#			***  None  ***

#	(4)	Change history:
#			***  None  ***

#	(5)	Definitions:

#			Namespaces:
#				Arena2SaveGame		--	Dialog manager for a picklist
#										of saved games

#			Exports:
#				saveGameDialog		--	Launch the dialog manager

#			Assertions about definitions:
#				The dialog manager must be opened using a Tcl/Tk
#				namespace `class' that exports the `invoke' method,
#				used when a saved game is selected by the user.

#				The widget also imports its color scheme from the
#				same namespace 'class', which must provide the
#				required color parameters.

#	(6)	References:

#			Library files:
#				***  None  ***

#			Imports:
#				imgtoppm.tcl		--	Tcl class promoting 8-bit
#										pseudocolor to truecolor
#				$owner::invoke		--	The method to activate when a
#										a saved gane is selected

#			Assertions about the references:
#				The Tcl class passed to saveGameDialog must implement
#				the method

#					proc class::canInvoke {saveId} {}
#					proc class::invoke {saveId title} {}

#				where saveId is the selected save game slot#
#				and title is the user's name for that save slot.

#				`canInvoke' answers true if a fresh instance for
#				saveId is necessary and false if an instance for
#				that saveId is already open.

#				`invoke' builds and displays the class widget hierarchy
#				for the specified saved game.


#								 ********
#								** Text **
#								 ********


#							 *****************
#							** Configuration **
#							 *****************

namespace eval Arena2SaveGame {
	variable applicationBackground	ShellBackground
	variable background				ShellBackground
	variable buttonActiveBackground	MenuTopShadow
	variable buttonActiveForeground	PaleTurquoise1
	variable buttonBackground		IconForeground
	variable buttonForeground		White
	variable checkColor				gold
	variable columnBackground		ShellBackground
	variable columnForeground		gold
	variable computeTrough			gold2
	variable copyright				"\xa9"
	variable daggerTrough			SteelBlue
	variable editBackground			MidnightBlue
	variable editDisabledBackground	SteelBlue
	variable editDisabledForeground	White
	variable editForeground			White
	variable foreground				White
	variable gamma					2.4
	variable icon					"dagger.gif"
	variable legendBackground		SteelBlue
	variable majorSkillColor		PaleTurquoise1
	variable minorSkillColor		gold
	variable owner					{}
	variable popupBackground		EditBackground
	variable popupForeground		MistyRose
	variable primarySkillColor		LightSkyBlue
	variable rowBackground			ShellBackground
	variable rowForeground			White
	variable sectionBackground		maroon4
	variable sectionForeground		White
	variable selectBackground		orange3
	variable selectForeground		#68f500000000
	variable trademark				"\xae"

	namespace export saveGameDialog
}
namespace import -force Arena2SaveGame::saveGameDialog

source imgtoppm.tcl
	namespace import -force Arena2Image::blankImage
	namespace import -force Arena2Image::buildPalette
	namespace import -force Arena2Image::buildPaletteFrom
	namespace import -force Arena2Image::convertImage
	namespace import -force Arena2Image::convertRaw

if {[llength [namespace children :: ::Arena2Config]] > 0} then {
	variable Arena2Config::artDir
	variable Arena2Config::exeDir
	variable Arena2Config::tmpDir
	variable Arena2Image::palette [file join "$Arena2Config::artDir" "oldpal.pal"]
	variable Arena2SaveGame::pal [file join "$Arena2Config::artDir" "art_pal.col"]
	variable Arena2SaveTree::tmpDir $Arena2Config::tmpDir
	variable Arena2SaveGame::exeDir $Arena2Config::exeDir
	variable Arena2SaveGame::artDir $Arena2Config::artDir
}

#							 ********************
#							** Instance methods **
#							 ********************

proc Arena2SaveGame::aboutBox {} {
#		Display the stock `about' box.

	variable copyright
	variable trademark
	if {$::Arena2Config::hasExtraFont} then {
		tk_messageBox \
			-font "-*-Helvetica-*-r-*-*-17-*-85-85-p-*-*-*" \
			-type ok \
			-title "About Chunk Viewer" \
			-message \
"     Daggerfall$trademark\n   Chunk Viewer\n\
\nCopyright($copyright) 2002\nNagual Enterprises"
	} else {
		tk_messageBox \
			-type ok \
			-title "About Chunk Viewer" \
			-message \
"     Daggerfall$trademark\n   Chunk Viewer\n\
\nCopyright($copyright) 2002\nNagual Enterprises"
	}
}

proc Arena2SaveGame::addSaveGame {index} {
#		Add the save game info for a saved game slot index to the
#		initial picklist display.

	variable ::Arena2SaveTree::exeDir
	variable ::Arena2SaveTree::tmpDir
	variable ::Arena2SaveTree::saveTree
	variable ::Arena2SaveTree::saveName
	variable ::Arena2SaveTree::saveImage
	variable gamma

	set ppmImage [file join $tmpDir "save$index.ppm"]
	set saveDir [file join $exeDir "save$index"]
	set savePath [file join $saveDir $saveTree]
	set imageForm "savetree$index"
	catch {image delete $imageForm}
	image create photo $imageForm -gamma $gamma
	if { [file exists $savePath] } then {
		set name [asString [read_file [file join $saveDir $saveName]]]
		set image [file join $saveDir $saveImage]
		Arena2Image::convertRaw $image $ppmImage 80 40
		lappend tag dagtree dagtree-save$index "status-$name"
	} else {
		set name "-none-"
		set image [file join $tmpDir "save$index.ppm"]
		Arena2Image::blankImage $image 80 40
		lappend tag dagtree
	}
	$imageForm configure -file $ppmImage
	.t insert end " " {dagtreespace}
	set ii [.t image create end -image $imageForm -padx 4 -pady 4]
	set lastLine [.t index $imageForm]
	foreach one $tag {
		.t tag add $one $ii
	}
	.t insert end " $name" $tag
	.t insert end " \n " {dagtreespace}
}

proc Arena2SaveGame::asString {bytes} {
#		Answer the byte stream terminated at the first NUL.
#		The ordinary full binary scan isn't portable between
#		Os platforms.  In particular, the Windows port is
#		broken.

	set size [clength "$bytes"]
	set index 0
	while {$index < $size} {
		set byte [crange "$bytes" $index $index]
		binary scan $byte c code
		if {$code == 0} then {
			return [crange "$bytes" 0 [expr $index - 1]]
		}
		incr index
	}
	return "$bytes"
}

proc Arena2SaveGame::close { w } {
#		Destroy the application.

#	Clear the debris from the animator.  The file cmd is dim-witted,
#	so avoid giving it an empty list to process.

	set blob [concat \
			 [glob -nocomplain [file join $::ChunkView::tmpDir "save*p*.??"]] \
			 [glob -nocomplain [file join $::ChunkView::tmpDir "save*.ppm"]]]
	if {[llength $blob] > 0} then {
		eval file delete $blob
	}
#	Finally, remove the ensemble of widgets from the screen.

	destroy $w
}

proc Arena2SaveGame::invoke index {
#		Given the user-selected save game index, open an editor on the
#		underlying savetree.dat.

	variable owner
	set tags [.t tag names $index]
	set i [lsearch -glob $tags dagtree-save*]
	if {$i < 0} {
		return
	}
	set view [string range [lindex $tags $i] 8 end]
	if {![::$owner\::canInvoke $view]} then {
		return
	}
	set cursor [.t cget -cursor]
	.t configure -cursor watch
	update
	set i [lsearch -glob $tags status-*]
	set status [string range [lindex $tags $i] 7 end]
	eval ::$owner\::invoke $view "$status"
#	uplevel [list source $view.tcl]
	update
	.t configure -cursor $cursor
	.t tag add visited "$index linestart +1 chars" "$index lineend -1 chars"
}

proc Arena2SaveGame::saveGameDialog {class} {
#		Build the initial picklist for save game edits.

#	Inherit the color scheme from the specified namespace.

	eval set applicationBackground \$::$class\::applicationBackground
	eval set buttonActiveBackground \$::$class\::buttonActiveBackground
	eval set buttonActiveForeground \$::$class\::buttonActiveForeground
	eval set buttonBackground \$::$class\::buttonBackground
	eval set foreground \$::$class\::foreground
	eval set popupBackground \$::$class\::popupBackground
	eval set popupForeground \$::$class\::popupForeground
	eval set selectBackground \$::$class\::selectBackground
	eval set selectForeground \$::$class\::selectForeground
	variable gamma

	variable owner $class
	variable icon
	global tcl_platform

	menu	.menuBar \
			-tearoff 0 \
			-background $popupBackground \
			-foreground $popupForeground \
			-activebackground $selectBackground \
			-activeforeground $selectForeground \
			-disabledforeground $popupForeground \
			-activeborderwidth 1 \
			-borderwidth 1
	.menuBar add cascade \
			 -menu .menuBar.file \
			 -label "File" \
			 -underline 0
	menu	.menuBar.file \
			-tearoff 0 \
			-background $popupBackground \
			-foreground $popupForeground \
			-activebackground $selectBackground \
			-activeforeground $selectForeground \
			-disabledforeground $popupForeground

	if {$tcl_platform(platform) == "macintosh"} {
#		On the Mac use the special .apple menu for the about item

		.menuBar add cascade \
				-menu .menuBar.apple
		menu	.menuBar.apple \
				-tearoff 0
		.menuBar.apple add command \
				-label "About..." \
				-command "Arena2SaveGame::aboutBox"
	} else {
		.menuBar.file add command \
				-label "About..." \
				-command "Arena2SaveGame::aboutBox" \
				-underline 0 -accelerator "F1" \
				-background $popupBackground \
				-foreground $popupForeground \
				-activebackground $selectBackground \
				-activeforeground $selectForeground
		.menuBar.file add sep \
				-background $popupBackground
	}
	.menuBar.file add command \
			-label "Quit" \
			-command "::Arena2SaveGame::close ." \
			-underline 0 \
			-accelerator "Alt-X" \
			-background $popupBackground \
			-foreground $popupForeground \
			-activebackground $selectBackground \
			-activeforeground $selectForeground
	. configure -menu .menuBar
	bind . <F1> Arena2SaveGame::aboutBox
	bind . <Alt-Key-x> "Arena2SaveGame::close ."
#	catch {image delete tes2Icon}
#	image create photo tes2Icon
#	tes2Icon configure -file $icon
#	button .b -image tes2Icon
#	wm iconwindow . .b

	frame	.statusBar \
			-background $applicationBackground
	label	.statusBar.lab \
			-text "   " \
			-relief sunken \
			-bd 1 \
			-font -*-Helvetica-Bold-R-Normal--*-120-*-*-*-*-*-* \
			-anchor w \
			-background $applicationBackground \
			-foreground $foreground
	label	.statusBar.foo \
			-width 8 \
			-relief sunken \
			-bd 1 \
			-font -*-Helvetica-Bold-R-Normal--*-120-*-*-*-*-*-* \
			-anchor w \
			-background $applicationBackground \
			-foreground $foreground
	pack	.statusBar.lab \
			-side left \
			-padx 2 \
			-expand yes \
			-fill both
	pack	.statusBar.foo \
			-side left \
			-padx 2
	pack	.statusBar \
			-side bottom \
			-fill x \
			-pady 2

	frame	.textFrame \
			-background $applicationBackground
	scrollbar .s \
			-orient vertical \
			-command {.t yview} \
			-highlightthickness 0 \
			-takefocus 1 \
			-background $applicationBackground \
			-activebackground $buttonActiveBackground \
			-troughcolor $buttonBackground \
			-width 12
	pack	.s \
			-in .textFrame \
			-side right \
			-fill y
	text	.t \
		 	-yscrollcommand {.s set} \
		 	-wrap word \
		 	-width 40 \
		 	-height 17 \
		 	-font {Helvetica 13} \
		 	-setgrid 1 \
		 	-highlightthickness 0 \
		 	-padx 4 \
		 	-pady 2 \
		 	-takefocus 0 \
		 	-background $applicationBackground \
		 	-foreground $foreground
	pack	.t \
			-in .textFrame \
			-expand y \
			-fill both \
			-padx 1
	pack	.textFrame \
			-expand yes \
			-fill both

#	Create a bunch of tags to use in the text widget, such as those for
#	section titles and dagtree descriptions.  Also define the bindings for
#	tags.

	.t tag configure title \
		-font {Helvetica 14 bold} \
		-justify center
	.t tag configure subtitle \
		-justify left

#	Put some "space" characters to the left and right of each browser
#	description so they are highlighted only when the mouse cursor
#	is right over them
	
	.t tag configure dagtreespace -lmargin1 1c -lmargin2 1c

	if {[winfo depth .] == 1} {
		.t tag configure dagtree \
			-lmargin1 1c \
			-lmargin2 1c \
			-underline 1
		.t tag configure visited \
			-lmargin1 1c \
			-lmargin2 1c \
			-underline 1
		.t tag configure hot \
			-background black \
			-foreground white
	} else {
		.t tag configure dagtree \
			-lmargin1 1c \
			-lmargin2 1c \
			-foreground "#dbd7b4" \
			-underline 0
		.t tag configure visited \
			-lmargin1 1c \
			-lmargin2 1c \
			-foreground "#b0d0ff" \
			-underline 0
		.t tag configure hot \
			-foreground yellow \
			-background $buttonActiveBackground \
			-underline 0
	}
	.t tag bind dagtree <ButtonRelease-1> \
		{Arena2SaveGame::invoke [.t index {@%x,%y}]}
	set lastLine ""
	.t tag bind dagtree <Enter> {
		set lastLine [.t index {@%x,%y linestart}]
		.t tag add hot "$lastLine +1 chars" "$lastLine lineend -1 chars"
		.t config -cursor hand2
		Arena2SaveGame::showStatus [.t index {@%x,%y}]
	}
	.t tag bind dagtree <Leave> {
		.t tag remove hot 1.0 end
		.t config -cursor xterm
		.statusBar.lab config -text ""
	}
	.t tag bind dagtree <Motion> {
		set newLine [.t index {@%x,%y linestart}]
		if {[string compare $newLine $lastLine] != 0} {
			.t tag remove hot 1.0 end
			set lastLine $newLine
			set tags [.t tag names {@%x,%y}]
			set i [lsearch -glob $tags dagtree-save*]
			if {$i >= 0} {
				.t tag add hot "$lastLine +1 chars" "$lastLine lineend -1 chars"
			}
		}
		Arena2SaveGame::showStatus [.t index {@%x,%y}]
	}
#	Create the text for the text widget.

	catch {image delete dagger}
	image create photo dagger -file dagger.gif

#	.t insert end "Daggerfall\[tm\] Character SaveGame" title
	.t insert end "Daggerfall\xae Saved Game" title
	.t image create end -image dagger -padx 4 -pady 4
	.t insert end " \n " {dagtreespace}
	busy hold .
	update
	set count 0
	while {$count < 6} {
		Arena2SaveGame::addSaveGame $count
		incr count
	}
	.t configure -state disabled
	busy release .
	focus .s
}

proc Arena2SaveGame::showStatus index {
#		Show the name of a save game in the status bar. This procedure
#		is called when the user moves the cursor over a saved game in
#		the list of available games to edit.

	set tags [.t tag names $index]
	set i [lsearch -glob $tags status-*]
	set cursor [.t cget -cursor]
	if {$i < 0} {
		.statusBar.lab config -text " "
		set newcursor xterm
	} else {
		set viewer [string range [lindex $tags $i] 7 end]
		.statusBar.lab config -text "Display $viewer"
		set newcursor hand2
	}
	if [string compare $cursor $newcursor] {
		.t config -cursor $newcursor
	}
}

#eval destroy [winfo child .]
#wm title . "TES2 Saved Games"
#Arena2Image::buildPalette
#Arena2SaveGame::saveGameDialog ChunkView
